#include "LecteurRedacteur.hpp"
#include <ostream>
#include <sstream>

using namespace std;

LecteurRedacteur::LecteurRedacteur() :
  nb_lecteurs(0), nb_redacteurs(0),nb_redacteurs_bloques(0),
  mut(), mut_ordre(),
  cond_lecteur(), cond_redacteur()
{
}


LecteurRedacteur::~LecteurRedacteur() {
}

void
LecteurRedacteur::DebutLecture() {
  mut.lock();
  while((nb_redacteurs>0)||(nb_redacteurs_bloques >0)) {
    cond_lecteur.wait(mut);
  }
  nb_lecteurs++;
  mut.unlock();
}


void
LecteurRedacteur::FinLecture() {
  mut.lock();
  nb_lecteurs--;
  if (nb_lecteurs == 0) {
    cond_redacteur.notify_one();
  }
  mut.unlock();
}

void
LecteurRedacteur::DebutRedaction() {
  mut.lock();
  // Attente "polie"
  if((nb_redacteurs>0)||(nb_lecteurs>0)) {
    // Petit manque de rigueur (acceptable d'après l'énoncé) : si le
    // wait_for est réveillé avant l'expiration du timeout, le code
    // passe directement à la boucle suivante. L'attente "polie" a
    // duré moins d'une seconde. On pourrait s'en sortir avec
    // wait_until.
    cond_redacteur.wait_for(mut, chrono::seconds(1));
  }
  nb_redacteurs_bloques++;
  // À partir de maintenant, les lecteurs ne peuvent plus accéder à la
  // section critique.
  while((nb_redacteurs>0)||(nb_lecteurs>0)) {
    cout << "thread bloque" << endl;
    cond_redacteur.wait(mut);
  }
  nb_redacteurs_bloques--;
  nb_redacteurs++;
  mut.unlock();
}

void
LecteurRedacteur::FinRedaction() {
  mut.lock();
  cond_redacteur.notify_one();
  if (nb_redacteurs_bloques==0) {
    cond_lecteur.notify_all();
  }
  nb_redacteurs--;
  mut.unlock();
}

ostream& operator<<(ostream &out, const LecteurRedacteur& lr) {
  ostringstream c;
  c << "nblect=" << lr.nb_lecteurs
    << "\tnbred=" << lr.nb_redacteurs
    << "\tnbredbloq=" << lr.nb_redacteurs_bloques;

  out << c.str();
  return out;
}
