#include <thread>
#include <mutex>
#include <iostream>
#include <vector>
#include <condition_variable>
#include <unistd.h>
#include <climits>

#define MAX_ETUDIANTS 100
#define MAX_RENDUS 20
using namespace std;

/**
 * Fonction appelée là où vous devez écrire du code.
 * Ne modifiez pas la fonction, mais vous ne devriez
 * plus l'appeler si vous terminez le TP.
 */
#define A_FAIRE() do { cout << "Ligne " << __LINE__ << ": Cette partie n'est pas encore faite, à vous !" << endl; exit(1); } while (0)

// Vous pouvez modifier cette constante pour vous
// aider à debugger
const bool debug = false;

/**
 * Classe de base commune aux moniteurs.
 * Ne pas modifier.
 */
class Plateforme
{
public:
  virtual bool deposerRendu(int) = 0;
  virtual void finaliserRendus() = 0;
  virtual void archiverRendus() = 0;

protected:
  void depotEffectif()
  {
    unique_lock<mutex> l(m);
    nb_rendus_non_sauvegardes++;
    if (debug) {
      cout << "depotEffectif()" << endl;
    }
    verification();
  }

  void archiverRendusEffectif()
  {
    unique_lock<mutex> l(m);
    nb_rendus_non_sauvegardes = 0;
    if (debug)
    {
      cout << "archiverRendusEffectif()" << endl;
    }
    verification();
  }

private:
  void verification()
  {
    if (nb_rendus_non_sauvegardes > MAX_RENDUS)
      {
	cout << "Erreur trop de rendus" << endl;
	exit(2);
      }
  }
  mutex m;
  int nb_rendus_non_sauvegardes = 0;
};

Plateforme *gestion_plateforme = nullptr;
int nb_etudiants = 6;
int version;
bool encore_des_rendus = true;

// Pour version equitable
int deja_rendus[MAX_ETUDIANTS];
int veut_rendre[MAX_ETUDIANTS];

/**
 * Version qui ne fait rien.
 * Ne pas modifier, sauf le commentaire À_FAIRE.
 */
class Plateforme_Sans_Protection : public Plateforme
{
  bool deposerRendu(int id) {
    nb_rendus_non_sauvegardes++;
    nb_rendus_total++;
    depotEffectif();
    return true;
  };

  void archiverRendus(){
    archiverRendusEffectif();
    // On reinitialise le nombre de rendus que l'on
    // peut faire avant traitement par l'enseignant
    nb_rendus_non_sauvegardes = 0;
  }
  
  void finaliserRendus(){
    encore_des_rendus = false;
    cout << "Le nombre de rendus finalisés est "
	 << nb_rendus_total << endl;
  }

  /* À FAIRE
     Expliquez le problème avec cette classe (cf. énoncé).
     Répondez ici :





  */

private:
  int nb_rendus_non_sauvegardes = 0;
  int nb_rendus_total           = 0;
};

class Plateforme_Sur : public Plateforme
{
  // À FAIRE : gérer correctement l'accès concurrent
  // à la plateforme. Il est interdit de supprimer
  // les appels aux fonctions depotEffectif() et
  // deposerRendusEffectif() mais vous ne devez appeler
  // depotEffectif() que lorsque le dépôt est
  // effectivement réalisé, c'est à dire quand la
  // fonction renvoie 'true'
  bool deposerRendu(int id) {
    bool got_place = false;
    m.lock();
    if (nb_rendus_non_sauvegardes < MAX_RENDUS){
      nb_rendus_non_sauvegardes++;
      nb_rendus_total++;
      got_place = true;
      depotEffectif();
    }
    m.unlock();
    return got_place;
  };

  void archiverRendus(){
    m.lock();
    archiverRendusEffectif();
    nb_rendus_non_sauvegardes = 0;
    m.unlock();
  }

  void finaliserRendus(){
    m.lock();
    encore_des_rendus = false;
    cout << "Le nombre de rendus finalisés est "
	 << nb_rendus_total << endl;
    m.unlock();
  }
private:
  int nb_rendus_non_sauvegardes = 0;
  int nb_rendus_total           = 0;
  // À FAIRE : ajouter le nécessaire pour que la classe
  // se comporte comme un moniteur de Hoare.
  mutex m;
};

class Plateforme_Avec_Attente : public Plateforme
{
  // À FAIRE : gérer correctement l'accès concurrent
  // à la plateforme. Il est interdit de supprimer
  // les appels aux fonctions depotEffectif() et
  // deposerRendusEffectif() mais vous ne devez appeler
  // depotEffectif() que lorsque le dépôt est
  // effectivement réalisé, c'est à dire quand la
  // fonction renvoie 'true'
  bool deposerRendu(int id) {
    m.lock();

    // Version minimale sans la gestion de encore_des_rendus
    // Attention avec le cas encore_des_rendus à false il faut bien faire un notify dans finaliserRendus() !
    while (nb_rendus_non_sauvegardes >= MAX_RENDUS && encore_des_rendus){
      c.wait(m);
    }
    // Invariant : nb_rendus_non_sauvegardes < MAX_RENDUS || !encore_des_rendus
    if (!encore_des_rendus){
      m.unlock();
      return false;
    }
    // Invariant : nb_rendus_non_sauvegardes < MAX_RENDUS
    nb_rendus_non_sauvegardes++;
    nb_rendus_total++;
    depotEffectif();
    m.unlock();
    return true;
  };

  void archiverRendus(){
    m.lock();
    archiverRendusEffectif();
    nb_rendus_non_sauvegardes = 0;
    c.notify_all();
    m.unlock();
  }

  void finaliserRendus(){
    m.lock();
    encore_des_rendus = false;
    c.notify_all();
    cout << "Le nombre de rendus finalisés est "
	 << nb_rendus_total << endl;
    m.unlock();
  }

private:
  int nb_rendus_non_sauvegardes = 0;
  int nb_rendus_total           = 0;
  // À FAIRE : ajouter le nécessaire pour que la classe
  // se comporte comme un moniteur de Hoare.
  mutex m;
  condition_variable_any c;
};

int min_rendus(){
  int min_des_rendus = INT_MAX;
  for (int i=0 ; i <nb_etudiants; i++){
    if (veut_rendre[i] && (deja_rendus[i] < min_des_rendus)){
      min_des_rendus = deja_rendus[i];
    }
  }
  return min_des_rendus;
}

class Plateforme_Equitable : public Plateforme
{
  // À FAIRE : gérer correctement l'accès concurrent
  // à la plateforme. Il est interdit de supprimer
  // les appels aux fonctions depotEffectif() et
  // deposerRendusEffectif() mais vous ne devez appeler
  // depotEffectif() que lorsque le dépôt est
  // effectivement réalisé, c'est à dire quand la
  // fonction renvoie 'true'
  bool deposerRendu(int id) {
    m.lock();
    veut_rendre[id]++;
    while ((nb_rendus_non_sauvegardes >= MAX_RENDUS || deja_rendus[id] != min_rendus())
	   &&
	   encore_des_rendus){
      c.wait(m);
    }
    // Invariant : nb_rendus_non_sauvegardes < MAX_RENDUS && deja_rendus[id] == min_rendus() || !encore_des_rendus
    if (!encore_des_rendus)
    {
      m.unlock();
      return false;
    }
    // Invariant : nb_rendus_non_sauvegardes < MAX_RENDUS && deja_rendus[id] == min_rendus()
    veut_rendre[id]--;
    deja_rendus[id]++;
    nb_rendus_non_sauvegardes++;
    nb_rendus_total++;
    depotEffectif();
    m.unlock();
    return true;
  };

  void archiverRendus(){
    m.lock();
    archiverRendusEffectif();
    nb_rendus_non_sauvegardes = 0;
    c.notify_all();
    m.unlock();
  }

  void finaliserRendus(){
    encore_des_rendus = false;
    c.notify_all();
    cout << "Le nombre de rendus finalisés est "
	 << nb_rendus_total << endl;
  }
private:
  int nb_rendus_non_sauvegardes = 0;
  int nb_rendus_total           = 0;
  // À FAIRE : ajouter le nécessaire pour que la classe
  // se comporte comme un moniteur de Hoare.
  mutex m;
  condition_variable_any c;
};

  /* À FAIRE
     Justifier qu'il n'y a plus de problème d'équitabilité
     avec votre approche.



  */



/**
 * Comportement d'un thread "etudiant"
 * Ne pas modifier.
 */
void etudiant(int id, int nb_rendus)
{
  int nb_rendus_reussis = 0;
  for (int i = 0; i<nb_rendus; i++){
    int ok = gestion_plateforme->deposerRendu(id); // On dépose régulièrement
    nb_rendus_reussis += ok;
  }

  cout << "Etudiant ID " << id << " a réussi "
       << nb_rendus_reussis << " rendus" << endl;
}

/**
 * Comportement d'un thread "enseignant"
 * Ne pas modifier.
 */
void enseignant()
{
  while(encore_des_rendus){
    usleep(400); // On attend
    gestion_plateforme->archiverRendus();
  }

  cout << "Enseignant terminé !" << endl;
}

void enseignant_limite()
{
  int nb = 1000;
  while(nb > 0){
    gestion_plateforme->archiverRendus();
    nb--;
  }
  gestion_plateforme->finaliserRendus(); // indice, dur a trouver

  cout << "Enseignant terminé !" << endl;
}

void lancement_multithread()
{
  vector<thread> tab;
  int nb_rendus[MAX_ETUDIANTS];

  int rendus_attendus=0;

  // Initialisation du nombre de rendus qu'essaiera de
  // faire chaque étudiant
  for (int i = 0; i < nb_etudiants; i++){
    nb_rendus[i] = rand()%100;
    rendus_attendus += nb_rendus[i];
    cout << "Etudiant ID " << i << " fera "
	 << nb_rendus[i] << " rendus" << endl;
  }
  cout << rendus_attendus
       << " rendus sont attendus" << endl;

  // À FAIRE : lancer nb_etudiants threads "etudiants" et
  // un thread "enseignant", puis attendre que les threads
  // aient terminé.
  
  for (int i = 0; i < nb_etudiants; i++){
    tab.push_back(thread(etudiant, i, nb_rendus[i]));
  }
  thread t_enseignant;
  if (version <= 2) {
    t_enseignant = thread(enseignant);
  } else {
    t_enseignant = thread(enseignant_limite);
  }
		     
  for (int i = 0; i < nb_etudiants; i++){
    tab[i].join();
  }
  gestion_plateforme->finaliserRendus();  
  t_enseignant.join();

}

/**
 * Fonction principale qui analyse la ligne de commande.
 * Ne pas modifier.
 */
int main(int argc, char **argv)
{
  cout << "Entrez le nombre d'étudiants (max 6)"
       << endl;
  cin  >> nb_etudiants;

  if (nb_etudiants > 6){
    cout << "Au maximum 6 étudiants" << endl;
    return 1;
  }
  
  cout << "Il y a " << nb_etudiants
       << " etudiants" << endl;

  cout << "Entrez la version du code: " << endl
       << "0: sans protection (original)" << endl
       << "1: sur (solution naïve)" << endl
       << "2: avec attente" << endl
       << "3: avec attente (gestion de la fin de l'examen)" << endl
       << "4: equitable" << endl;
  cin >> version;

  switch (version){
  case 0:
    cout << "Version sans protection (original)" << endl;
    gestion_plateforme = new Plateforme_Sans_Protection();
    break;
  case 1:
    cout << "Version avec protection (naïve)" << endl;
    gestion_plateforme = new Plateforme_Sur();
    break;
  case 2:
    cout << "Version avec attente" << endl;
    gestion_plateforme = new Plateforme_Avec_Attente();
    break;
  case 3:
    cout << "Version avec attente (gestion de la fin de l'examen)" << endl;
    gestion_plateforme = new Plateforme_Avec_Attente();
    break;
  case 4:
    cout << "Version équitable" << endl;
    gestion_plateforme = new Plateforme_Equitable();
    break;
  default:
    cout << "ERROR: Unknown option: " << endl;
    return 1;
  }
  lancement_multithread();
  return 0;
}

// L'équivalent d'une plateforme par étudiant. Tableau de rendus et tableau de mutex / variables de condition

/* À FAIRE
   Dans notre approche les étudiants déposent tous leurs
   rendus au même endroit. Que faudrait-il faire si 
   chaque étudiant avait sa propre plateforme de dépôt ?
   
   
   
   
*/

