// À COMPLETER : écrire ici votre nom et numéro de machine :
// Nom : ...
// Prénom : ...
// Numéro étudiant : ...
// Numéro de machine : ...
/********************************************************************
 * Copyright (C) 2023 by UCBL                                       *
 * Initial author: Gregoire Pichon                                  *
 ********************************************************************/

#include <thread>
#include <mutex>
#include <iostream>
#include <vector>
#include <condition_variable>
#include <unistd.h>
#include <climits>

#define MAX_ETUDIANTS 100
#define MAX_RENDUS 20
using namespace std;

/**
 * Fonction appelée là où vous devez écrire du code. Ne modifiez pas la
 * fonction, mais vous ne devriez plus l'appeler si vous terminez le TP.
 */
#define A_FAIRE() do {							\
    cout << "Ligne " << __LINE__ << ": Cette partie n'est pas encore faite, à vous !" << endl; \
    abort();								\
  } while (0)

const bool debug = true;
const bool check = true;

/**
 * Classe de base commune aux moniteurs.
 * Ne pas modifier.
 */
class Serveur
{
public:
  virtual bool faireDepot(int) = 0;
  virtual void finaliserDepots() = 0;
  virtual void sauvegarderDepots() = 0;

protected:
  void depotEffectif()
  {
    unique_lock<mutex> l(m);
    nb_depots_non_sauvegardes++;
    verification();
  }

  void sauvegardeEffective()
  {
    unique_lock<mutex> l(m);
    nb_depots_non_sauvegardes--;
    verification();
  }

private:
  void verification()
  {
    if (check && nb_depots_non_sauvegardes > nb_depots_max)
      {
	cout << "Erreur trop de depots" << endl;
	abort();
      }
  }
  mutex m;
  int nb_depots_non_sauvegardes     = 0;
  int nb_depots_max = MAX_RENDUS;
};

Serveur *gestion_serveur = nullptr;
int nb_etudiants = 8;
int version;
bool encore_des_depots = true;

// Pour famine
int deja_depots[MAX_ETUDIANTS];
int veut_rendre[MAX_ETUDIANTS];

/**
 * Version qui ne fait rien.
 * Ne pas modifier, sauf le commentaire À FAIRE.
 */
class Serveur_Sans_Protection : public Serveur
{
  bool faireDepot(int id) {
    nb_depots++;
    nb_depots_total++;
    depotEffectif();
    return true;
  };

  void sauvegarderDepots(){
    for (int i = 0; i < nb_depots; i++){
      sauvegardeEffective();
    }
    // On reinitialise le nombre de depots que l'on peut faire avant traitement par l'enseignant
    nb_depots = 0;
  }
  
  void finaliserDepots(){
    encore_des_depots = false;
    if (debug) {
      cout << "Le nombre de depots finalisés est " << nb_depots_total << endl;
    }
  }

  /* À FAIRE
     Expliquez le problème avec cette classe (cf. énoncé).
     Répondez ici :





  */

private:
  int nb_depots       = 0;
  int nb_depots_max   = MAX_RENDUS;
  int nb_depots_total = 0;
};

class Serveur_Sur : public Serveur
{
  // À FAIRE : gérer correctement l'accès concurrent au serveur
  // (Il est interdit de supprimer les appels aux fonctions depotEffectif()
  // et sauvegardeEffectuve() mais vous ne devez appeler depotEffectif() que
  // lorsque le dépôt est effectivement réalisé, c'est à dire quand la fonction
  // renvoie 'true')
  bool faireDepot(int id) {
    bool got_place = false;
    m.lock();
    if (nb_depots < nb_depots_max){
      nb_depots++;
      nb_depots_total++;
      got_place = true;
      depotEffectif();
    }
    m.unlock();
    return got_place;
  };

  void sauvegarderDepots(){
    m.lock();
    for (int i = 0; i < nb_depots; i++){
      sauvegardeEffective();
    }
    nb_depots = 0;
    m.unlock();
  }

  void finaliserDepots(){
    encore_des_depots = false;
    if (debug) {
      cout << "Le nombre de depots finalisés est " << nb_depots_total << endl;
    }
  }
private:
  // À FAIRE : ajouter le nécessaire pour que la classe se comporte comme un moniteur de Hoare.
  int nb_depots       = 0;
  int nb_depots_max   = MAX_RENDUS;
  int nb_depots_total = 0;
  mutex m;
};

class Serveur_Avec_Attente : public Serveur
{
  // À FAIRE : gérer correctement l'accès concurrent au serveur
  // (Il est interdit de supprimer les appels aux fonctions depotEffectif()
  // et sauvegardeEffective() mais vous ne devez appeler depotEffectif() que
  // lorsque le dépôt est effectivement réalisé, c'est à dire quand la fonction
  // renvoie 'true')
  bool faireDepot(int id) {
    m.lock();

    // Version minimale sans la gestion de encore_des_depots
    // Attention avec le cas encore_des_depots à false il faut bien faire un notify dans finaliserDepots() !
    while (nb_depots >= nb_depots_max && encore_des_depots){
      c.wait(m);
    }
    // Invariant : nb_depots < nb_depots_max || !encore_des_depots
    if (!encore_des_depots){
      m.unlock();
      return false;
    }
    // Invariant : nb_depots < nb_depots_max
    nb_depots++;
    nb_depots_total++;
    depotEffectif();
    m.unlock();
    return true;
  };

  void sauvegarderDepots(){
    m.lock();
    for (int i = 0; i < nb_depots; i++){
      sauvegardeEffective();
    }
    nb_depots = 0;
    c.notify_all();
    m.unlock();
  }

  void finaliserDepots(){
    encore_des_depots = false;
    c.notify_all();
    if (debug) {
      cout << "Le nombre de depots finalisés est " << nb_depots_total << endl;
    }
  }
private:
  // À FAIRE : ajouter le nécessaire pour que la classe se comporte comme un moniteur de Hoare.
  int nb_depots       = 0;
  int nb_depots_max   = MAX_RENDUS;
  int nb_depots_total = 0;
  mutex m;
  condition_variable_any c;
};

int min_depots(){
  int min_des_depots = INT_MAX;
  for (int i=0 ; i <nb_etudiants; i++){
    if (veut_rendre[i] && (deja_depots[i] < min_des_depots)){
      min_des_depots = deja_depots[i];
    }
  }
  return min_des_depots;
}

class Serveur_Sans_Famine : public Serveur
{
  // À FAIRE : gérer correctement l'accès concurrent au serveur
  // (Il est interdit de supprimer les appels aux fonctions depotEffectif()
  // et sauvegardeEffectuve() mais vous ne devez appeler depotEffectif() que
  // lorsque le dépôt est effectivement réalisé, c'est à dire quand la fonction
  // renvoie 'true')
  bool faireDepot(int id) {
    m.lock();
    veut_rendre[id]++;
    while ((nb_depots >= nb_depots_max || deja_depots[id] != min_depots())
	   &&
	   (encore_des_depots)){ // && nb_depots < nb_depots_max)){
      c.wait(m);
    }
    if (!encore_des_depots){
      m.unlock();
      return false;
    }
    veut_rendre[id]--;
    deja_depots[id]++;
    nb_depots++;
    nb_depots_total++;
    depotEffectif();
    m.unlock();
    return true;
  };

  void sauvegarderDepots(){
    m.lock();
    for (int i = 0; i < nb_depots; i++){
      sauvegardeEffective();
    }
    nb_depots = 0;
    c.notify_all();
    m.unlock();
  }

  void finaliserDepots(){
    encore_des_depots = false;
    c.notify_all();
    if (debug) {
      cout << "Le nombre de depots finalisés est " << nb_depots_total << endl;
    }
  }
private:
  // À FAIRE : ajouter le nécessaire pour que la classe se comporte comme un moniteur de Hoare.
  int nb_depots       = 0;
  int nb_depots_max   = MAX_RENDUS;
  int nb_depots_total = 0;
  mutex m;
  condition_variable_any c;
};

  /* À FAIRE
     Justifier qu'il n'y a plus de famine possible avec votre approche.



  */



/**
 * Comportement d'un thread "etudiant"
 */
void etudiant(int id, int nb_depots)
{
  int nb_depots_reussis = 0;
  for (int i = 0; i<nb_depots; i++){
    usleep(10%rand()%10);                          // On réfléchit
    int ok = gestion_serveur->faireDepot(id); // On dépose régulièrement
    nb_depots_reussis += ok;
  }

  cout << "Etudiant ID " << id << " a réussi " << nb_depots_reussis << " depots" << endl;
}

void enseignant()
{
  while(encore_des_depots){
    gestion_serveur->sauvegarderDepots();
  }
}

void enseignant_famine()
{
  int nb = 1000;
  while(nb > 0){
    gestion_serveur->sauvegarderDepots();
    nb--;
  }
  gestion_serveur->finaliserDepots(); // indice, dur a trouver

  if (debug) {
    cout << "Enseignant terminé !" << endl;
  }
}

void lancement_multithread()
{
  vector<thread> tab;
  int nb_depots[nb_etudiants];

  int depots_attendus=0;

  // Initialisation du nombre de depots qu'essaiera de faire chaque étudiant
  for (int i = 0; i < nb_etudiants; i++){
    nb_depots[i] = rand()%100;
    depots_attendus += nb_depots[i];
    cout << "Etudiant ID " << i << " fera " << nb_depots[i] << " depots" << endl;
  }
  cout << depots_attendus << " depots sont attendus" << endl;

  // À FAIRE : lancer nb_etudiants threads "etudiants" et un
  // thread "enseignant", puis attendre que les threads aient terminé.
  
  for (int i = 0; i < nb_etudiants; i++){
    tab.push_back(thread(etudiant, i, nb_depots[i]));
  }
  // thread t_enseignant = thread(enseignant);
  thread t_enseignant = thread(enseignant_famine);
		     
  for (int i = 0; i < nb_etudiants; i++){
    tab[i].join();
  }
  gestion_serveur->finaliserDepots();  
  t_enseignant.join();

}

/**
 * Fonction principale qui analyse la ligne de commande.
 * Ne pas modifier.
 */
int main(int argc, char **argv)
{
  gestion_serveur = new Serveur_Sans_Protection();
  
  cout << "Entrez le nombre d'étudiants (max 8)" << endl;
  cin  >> nb_etudiants;
  
  cout << "Il y a " << nb_etudiants << " etudiants" << endl;

  cout << "Entrez la version du code (0: sans protection, 1: sur, 2: avec attente, 3: sans famine)" << endl;
  cin >> version;

  switch (version){
  case 0:
    cout << "Version sans protection" << endl;
    gestion_serveur = new Serveur_Sans_Protection();
    break;
  case 1:
    cout << "Version avec protection" << endl;
    gestion_serveur = new Serveur_Sur();
    break;
  case 2:
    cout << "Version avec attente active" << endl;
    gestion_serveur = new Serveur_Avec_Attente();
    break;
  case 3:
    cout << "Version sans famine" << endl;
    gestion_serveur = new Serveur_Sans_Famine();
    break;
  default:
    cout << "ERROR: Unknown option: " << endl;
    exit(1);
  }
  lancement_multithread();
}

// L'équivalent d'un serveur par étudiant. Tableau de depots et tableau de mutex / variables de condition

/* À FAIRE
   Dans notre approche les étudiants déposent tous leurs depots au même
   endroit. Que faudrait-il faire si chaque étudiant avait sa propre
   serveur de dépôt ?
   
   
   
   
*/

