#include <thread>
#include <mutex>
#include <iostream>
#include <vector>
#include <condition_variable>
#include <unistd.h>
#include <climits>

#define MAX_VEHICULES 6
#define NB_TOURS      10

int PIETON  = 0;
int VELO    = 1;
int VOITURE = 2;
char NAMES[3][256] = {"pieton", "velo", "voiture"};

using namespace std;

/**
 * Fonction appelee là où vous devez ecrire du code.
 * Ne modifiez pas la fonction, mais vous ne devriez
 * plus l'appeler si vous terminez le TP.
 */
#define A_FAIRE() do { cout << "Ligne " << __LINE__ << ": Cette partie n'est pas encore faite, à vous !" << endl; exit(1); } while (0)

// Vous pouvez modifier cette constante pour vous
// aider à debugger
const bool debug = false;

mutex m_print;

void print_entree(int type_vehicule){
  unique_lock<mutex> l(m_print);
  cout << "Tentative entree de type "
       << NAMES[type_vehicule] << endl;
}

void print_sortie(int type_vehicule){
  unique_lock<mutex> l(m_print);
  cout << "Sortie de type "
       << NAMES[type_vehicule] << endl;
}

/**
 * Classe de base commune aux moniteurs.
 */
class Tunnel
{
public:
  virtual bool entreeTunnel(int) = 0;
  virtual void sortieTunnel(int) = 0;
  virtual int nb_attentes(void) = 0;

protected:
  void entreeEffective(int type_vehicule){
    unique_lock<mutex> l(m);

    // Les vélos ne passent pas s'il y a des piétons
    // Les voitures ne passent pas s'il y a
    // des piétons et/ou des vélos
    if ( (vehicules_entres[PIETON] > 0
	  && type_vehicule > PIETON)
	|| (vehicules_entres[VELO] > 0
	    && type_vehicule > VELO)){
      m_print.lock();
      cout << "Erreur passage non autorise" << endl;
      m_print.unlock();
      exit(2);      
    }
    
    vehicules_entres[type_vehicule]++;
    if (debug){
      m_print.lock();
	cout << "entreeEffective() type "
	     << NAMES[type_vehicule] << endl;
      m_print.unlock();
    }
  }

  void sortieEffective(int type_vehicule){
    unique_lock<mutex> l(m);
    vehicules_entres[type_vehicule]--;
    
    if ((vehicules_entres[PIETON] < 0)
	|| (vehicules_entres[VELO] < 0)
	|| (vehicules_entres[VOITURE] < 0)){
      m_print.lock();
      cout << "Erreur nombre négatif" << endl;
      m_print.unlock();
      exit(2);            
    }

    if (debug){
      m_print.lock();
      cout << "sortieEffective() type "
	   << NAMES[type_vehicule] << endl;
      m_print.unlock();
    }
  }

private:
  mutex m;
  int vehicules_entres[3] = {0, 0, 0};
};

Tunnel *gestion_tunnel = nullptr;
int nb_vehicules = 6;
int version;

class Tunnel_Sur : public Tunnel
{
  // À FAIRE : gerer correctement l'accès concurrent
  // au tunnel. Il est interdit de supprimer
  // les appels aux fonctions entreeEffective() et
  // sortieEffective()
  bool entreeTunnel(int type_vehicule) {
    unique_lock<mutex> l(m);
    bool success = true;
    if (debug){ print_entree(type_vehicule); }

    // Les vélos ne passent pas s'il y a des piétons
    // Les voitures ne passent pas s'il y a
    // des piétons et/ou des vélos
    if ( (vehicules_entres[PIETON] > 0
	  && type_vehicule > PIETON)
	|| (vehicules_entres[VELO] > 0
	    && type_vehicule > VELO)){
      success = false;
    }
    
    if (success){
      vehicules_entres[type_vehicule]++;
      entreeEffective(type_vehicule);
    }
    return success;
  };

  void sortieTunnel(int type_vehicule){
    unique_lock<mutex> l(m);
    if (debug){ print_sortie(type_vehicule); }
    vehicules_entres[type_vehicule]--;
    sortieEffective(type_vehicule);
  }

  // Pas utilisé dans cette version
  int nb_attentes(){
    return 0;
  }

private:
  mutex m;
  int vehicules_entres[3] = {0, 0, 0};
};

class Tunnel_Avec_Attente_Simple : public Tunnel
{
  // À FAIRE : gerer correctement l'accès concurrent
  // au tunnel avec attente passive. Il est interdit de
  // supprimer les appels aux fonctions
  // entreeEffective() et sortieEffective()
  bool entreeTunnel(int type_vehicule) {
    unique_lock<mutex> l(m);
    if (debug){ print_entree(type_vehicule); }

    while ((vehicules_entres[PIETON] > 0 && type_vehicule > PIETON)
	|| (vehicules_entres[VELO] > 0 && type_vehicule > VELO)){
      c_entree.wait(l);
      attentes++;
    }
    
    vehicules_entres[type_vehicule]++;
    entreeEffective(type_vehicule);
    return true;
  };

  void sortieTunnel(int type_vehicule){
    unique_lock<mutex> l(m);
    if (debug){ print_sortie(type_vehicule); }
    vehicules_entres[type_vehicule]--;
    sortieEffective(type_vehicule);
    c_entree.notify_all();
  }
  int nb_attentes(){
    return attentes;
  }

private:
  mutex m;
  condition_variable c_entree;
  int vehicules_entres[3] = {0, 0, 0};
  // à utiliser pour comparaison avec
  // classe Tunnel_Avec_Attente_Avancee,
  // pas utile avant
  int attentes = 0; 
};

class Tunnel_Avec_Attente_Avancee : public Tunnel
{
  // À FAIRE : gerer correctement l'accès concurrent
  // au tunnel avec attente passive en limitant les
  // réveils inutiles. Il est interdit de
  // supprimer les appels aux fonctions
  // entreeEffective() et sortieEffective()
  bool entreeTunnel(int type_vehicule) {
    unique_lock<mutex> l(m);
    if (debug){ print_entree(type_vehicule); }

    while ((vehicules_entres[PIETON] > 0 && type_vehicule > PIETON)
	|| (vehicules_entres[VELO] > 0 && type_vehicule > VELO)){
      if (type_vehicule == VELO){
	c_entree_velo.wait(l);
      }
      if (type_vehicule == VOITURE){
	c_entree_voiture.wait(l);
      }
      attentes++;
    }
    
    vehicules_entres[type_vehicule]++;
    entreeEffective(type_vehicule);
    return true;
  };

  void sortieTunnel(int type_vehicule){
    unique_lock<mutex> l(m);
    if (debug){ print_sortie(type_vehicule); }
    vehicules_entres[type_vehicule]--;
    sortieEffective(type_vehicule);
    if (type_vehicule == PIETON){
      // On reveille les velos si plus de pietons
      if (vehicules_entres[PIETON] == 0){
	c_entree_velo.notify_all();
	// Et les voitures si plus de velos
	if (vehicules_entres[VELO] == 0){
	  c_entree_voiture.notify_all();
	}
      }
    }
    if (type_vehicule == VELO){
      // On reveille les voitures si plus de velos
      if (vehicules_entres[VELO] == 0){
	c_entree_voiture.notify_all();
      }
    }
  }
  int nb_attentes(){
    return attentes;
  }
  
private:
  mutex m;
  condition_variable c_entree_velo;
  condition_variable c_entree_voiture;
  int vehicules_entres[3] = {0, 0, 0};
  // à utiliser pour comparaison avec
  // classe Tunnel_Avec_Attente_Simple
  int attentes = 0;
};

bool famine(int type_vehicule, int vehicules_entres[3], int vehicules_attentes[3]){

  // Pas de vehicule de ce type déjà présent
  if (vehicules_entres[type_vehicule] == 0){
    return false;
  }
  bool attente_bigger_v = false;
  for (int i=type_vehicule+1; i<3; i++){
    if (vehicules_attentes[i] > 0){
      attente_bigger_v = true;
      cout << NAMES[type_vehicule] << ": j'attends volontairement" << endl;
    }
  }
  return attente_bigger_v;
}

class Tunnel_Sans_Famine : public Tunnel
{
  // À FAIRE : gerer correctement l'accès concurrent
  // au tunnel avec attente passive et sans famine.
  // Il est interdit de supprimer les appels aux
  // fonctions entreeEffective() et sortieEffective()
  bool entreeTunnel(int type_vehicule) {
    unique_lock<mutex> l(m);
    if (debug){ print_entree(type_vehicule); }

    // On indique qu'on veut entrer
    vehicules_attentes[type_vehicule]++;

    while ((vehicules_entres[PIETON] > 0 && type_vehicule > PIETON)
	   || (vehicules_entres[VELO] > 0 && type_vehicule > VELO)
	   || famine(type_vehicule, vehicules_entres, vehicules_attentes)){
      c_entree.wait(l);
      attentes++;
    }
    cout << NAMES[type_vehicule] << ": je rentre" << endl;

    // On indique qu'on a réussi à entrer
    vehicules_attentes[type_vehicule]--;
    
    vehicules_entres[type_vehicule]++;
    entreeEffective(type_vehicule);
    return true;
  };

  void sortieTunnel(int type_vehicule){
    unique_lock<mutex> l(m);
    if (debug){ print_sortie(type_vehicule); }
    vehicules_entres[type_vehicule]--;
    sortieEffective(type_vehicule);
    // Il faut mieux reveiller quoiqu'il arrive sinon risque deadlock
    c_entree.notify_all();
  }
  int nb_attentes(){
    return attentes;
  }
  int attentes = 0;

private:
  mutex m;
  condition_variable c_entree;
  int vehicules_attentes[3] = {0, 0, 0};
  int vehicules_entres[3] = {0, 0, 0};
};

/**
 * Comportement d'un thread "vehicule"
 */
// id est le numéro du thread
// type_vehicule est 0: pieton; 1: velo; 2: voiture
void vehicule(int id, int type_vehicule, int &nb_entrees_out)
{
  int nb_entrees = 0;
  for (int i = 0; i<NB_TOURS; i++){
    bool success = gestion_tunnel->entreeTunnel(type_vehicule);
    if (success){
      nb_entrees++;
      usleep(100*rand()%5); // Temps de traversee
      gestion_tunnel->sortieTunnel(type_vehicule);
    }
  }

  m_print.lock();
  cout << "VEHICULE ID " << id << " de type "
       << NAMES[type_vehicule ] << " a reussi "
       << nb_entrees << " entrees"
       << " sur " << NB_TOURS << " tours" << endl;
  m_print.unlock();
  nb_entrees_out = nb_entrees;
}

void lancement_multithread()
{
  int types_vehicules[MAX_VEHICULES];
  int nb_total_entrees = 0;
  for (int i = 0; i < nb_vehicules; i++){
    // On initialise aléatoirement le type de véhicule
    // 0: pieton; 1: vélo: 2: voiture
    types_vehicules[i] = rand()%3;
  }

  // À FAIRE : lancer nb_vehicules "vehicules"

  vector<thread> tab;
  int tot_entrees[MAX_VEHICULES];
  for (int i = 0; i < nb_vehicules; i++){
    tot_entrees[i] = 0;
    tab.push_back(thread(vehicule, i, types_vehicules[i], std::ref(tot_entrees[i])));
  }
		     
  for (int i = 0; i < nb_vehicules; i++){
    tab[i].join();
    nb_total_entrees += tot_entrees[i];
  }

  m_print.lock();
  cout << "Nombre total entrées " << nb_total_entrees
       << " / maximum: " << nb_vehicules*NB_TOURS
       << endl;
  cout << "Nombre attentes "
       << gestion_tunnel->nb_attentes() << endl;

  m_print.unlock();

  (void)types_vehicules;
}

/**
 * Fonction principale qui analyse la ligne de commande.
 * Ne pas modifier.
 */
int main(int argc, char **argv)
{
  cout << "Entrez le nombre de vehicules (max 6)"
       << endl;
  cin  >> nb_vehicules;

  if (nb_vehicules > 6){
    cout << "Au maximum 6 vehicules" << endl;
    return 1;
  }

  cout << "Il y a " << nb_vehicules
       << " vehicules" << endl;

  cout << "Entrez la version du code: " << endl
       << "0: sur (solution naïve)" << endl
       << "1: avec attente simple" << endl
       << "2: avec attente avancée" << endl
       << "3: sans famine" << endl;
  cin >> version;

  switch (version){
  case 0:
    cout << "Version avec protection (naïve)" << endl;
    gestion_tunnel = new Tunnel_Sur();
    break;
  case 1:
    cout << "Version avec attente simple" << endl;
    gestion_tunnel = new Tunnel_Avec_Attente_Simple();
    break;
  case 2:
    cout << "Version avec attente avancee" << endl;
    gestion_tunnel = new Tunnel_Avec_Attente_Avancee();
    break;
  case 3:
    cout << "Version sans famine" << endl;
    gestion_tunnel = new Tunnel_Sans_Famine();
    break;
  default:
    cout << "ERROR: Unknown option: " << endl;
    return 1;
  }
  lancement_multithread();
  return 0;
}

/* À FAIRE
   Expliquez ici le problème que vous avez avec
   l'affichage et comment le résoudre.
   
   
   
   
   
*/

/* À FAIRE
   Expliquez ici l'intérêt de réduire le nombre
   d'attentes (et donc de notify)
   
   
   
   
   
*/

/* À FAIRE
   Expliquez ici le problème de famine
   
   
   
   
   
*/

