#include <unistd.h>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <sstream>
#include <iostream>
#include <vector>
#include <cstdio>
#include <getopt.h>

using namespace std;


mutex mut;

static struct option long_options[] =
  {
    {"with-gaston",  no_argument, 0, 'g'},
    {"help",  no_argument, 0, 'h'},
    {"with-fantasio",  no_argument, 0, 'f'},
    {"nb_bouteilles",  required_argument, 0, 'b'},
    {0, 0, 0, 0}
  };

void usage(const char *nomexec) {
  cerr << "usage : >" << nomexec
       << " [options]" << endl;
  cerr << "\t -h|--help : \t affiche cette aide"  << endl;
  cerr << "\t --nb-bouteille number : \t nombre de bouteilles et de bouchons" << endl;
  cerr << "\t -g|--with-gaston : \t avec gaston" << endl;
  cerr << "\t -f|--with-fantasio : \t avec fantasio"  << endl;
}

class cMaBlBdV {
  bool une_bouteille;
  bool un_bouchon;
  mutex m;
  condition_variable pret;
  condition_variable place;
  bool fini;
  
public:
  bool marteauAttend();
  void placeBouteille();
  void placeBouchon();
  void marteauTermine();
  void termine();
  
  cMaBlBdV() :
    une_bouteille(false),
    un_bouchon(false),
    m(),
    pret(),
    place(),
    fini(false)
  {
  }
};


bool cont = true;
bool gaston_est_la = false;

bool cMaBlBdV::marteauAttend() {
  unique_lock<mutex> ul(m);
  
  while((!(une_bouteille && un_bouchon)) && (!fini)) {
    // fprintf(stdout, "%d %d %d %d\n", une_bouteille, un_bouchon, fini, (!(une_bouteille && un_bouchon)) || (fini));
    pret.wait(ul);
  }
  if (fini)
    return true;
  return false;
}

void cMaBlBdV::marteauTermine() {
  unique_lock<mutex> ul(m);
  une_bouteille = false;
  un_bouchon = false;
  place.notify_all();
}

void cMaBlBdV::placeBouteille() {
  unique_lock<mutex> ul(m);
  while(une_bouteille) {
    place.wait(ul);
  }
  une_bouteille = true;

  pret.notify_one();
}

void cMaBlBdV::placeBouchon() {
  unique_lock<mutex> ul(m);
  while(un_bouchon) {
    place.wait(ul);
  }
  un_bouchon = true;

  pret.notify_one();
}

void vlan() {
  ostringstream s;
  if (gaston_est_la) {
    s << "Menfin !" << endl;
    //    sleep(1);
    //    exit(1);
  } else {
    s << "Vlan !" << endl;    
  }
  usleep(1000*(rand()%10));
  cout << s.str();  
}



void cMaBlBdV::termine() {
  unique_lock<mutex> ul(m);
  fini = true;
  pret.notify_one();
}

void marteau(cMaBlBdV &m, int &nb_bouchee) {
  nb_bouchee = 0;
  while(true) {
    bool fini = m.marteauAttend();
    if (fini)
      break;
    mut.lock();
    vlan();
    mut.unlock();
    nb_bouchee ++;
    m.marteauTermine();
  }
  usleep(10000);
  m.termine();
}

void fantasio() {
  bool jailemutex=false;
  while(cont) {
    if (jailemutex) {
      if (!gaston_est_la){
	jailemutex=false;
	mut.unlock();
      }
    } else {
      //      usleep(10000000);
      mut.lock();
      if (gaston_est_la) {
	jailemutex = true;
      } else {
	mut.unlock();
      }
    }
  }
}

void bouteilles(cMaBlBdV &m, int nb_bouteilles) {
  int i;
  
  for (i=0; i<nb_bouteilles; i++) {
    usleep(10000+1000*(rand()%10));
    m.placeBouteille();
    cout << "Une bouteille \n";
  }
  usleep(10000);
  m.termine();
}

void bouchons(cMaBlBdV &m, int nb_bouchons) {
  int i;
  for (i=0; i<nb_bouchons; i++) {
    usleep(10000+1000*(rand()%10));
    m.placeBouchon();
    cout << "Un bouchon \n";
  }
}

void gaston() {
  while(cont) {
    usleep(100000*(rand()%10));
    gaston_est_la = true;
    fprintf(stdout, "attention !\n");
    usleep(10000*(rand()%10));
    gaston_est_la = false;
    fprintf(stdout, "ouf\n");
  }
}


int main(int argc, char *argv[]) {
  int c;
  int opt_index;
  int nb = 100;
  int nb_bouchees = 0;
  vector<thread> th;
  vector<thread> thfoug;
  bool with_gaston = false;
  bool with_fantasio = false;
  
  
  while((c = getopt_long(argc, argv, "hgfb:",
			 long_options,
			 &opt_index
			 )) != -1) {
    switch (c) {
    case 'g' :
      with_gaston = true;
      break;
    case 'f' :
      with_fantasio = true;
      break;
    case 'h' :
      usage(argv[0]);
      exit(0);
      break;
    case 'b' :
      nb = atoi(optarg);
      break;
    default:
      cerr << "?? option inconue : " << c << endl;
      usage(argv[0]);
      exit(1);
      break;
    }
  }
  
  cMaBlBdV m;

  srand(time(NULL));
  
  if (with_gaston) {
    thfoug.push_back(thread(gaston));
  }
  if (with_fantasio) {
    thfoug.push_back(thread(fantasio));
  }
  th.push_back(thread(marteau, ref(m), ref(nb_bouchees)));
  th.push_back(thread(bouteilles, ref(m), nb));
  th.push_back(thread(bouchons, ref(m), nb));
  

  for (auto &t : th){
    t.join();
  }
  cont = false;

  for (auto &t : thfoug){
    t.join();
  }
  

  cout << "C'est terminé, il y a " << nb_bouchees << " bouteille de vin"<<endl;

  return 0;
}





