/********************************************************************
 * Copyright (C) 2018 by UCBL / LIP                                 *
 * Initial author: Matthieu Moy                                     *
 ********************************************************************/

#include <iostream>
#include <thread>
#include <mutex>
#include <vector>
#include <unistd.h>
#include <condition_variable>
#include <assert.h>
#include <queue>
#include <sstream>

using namespace std;
#define NB_VOITURES 5
#define NB_PLACES 1
#define NB_ESSAIS 2

class Parking {
public:
	enum cote {
		EST,
		OUEST,
		NB_COTES
	};
	void demande_entree(cote c) {
		unique_lock<mutex> l(m);
		attentes_cotes.push(c);
		while (not(nb_place_libre > 0 && attentes_cotes.front() == c)) {
			stringstream s;
			s << "Attente. Voitures en attente : "
			  << attentes_cotes.size()
			  << ", premiere cote " << attentes_cotes.front()
			  << ".\n";
			cout << s.str();
			conditions[c].wait(l);
		}
		attentes_cotes.pop();
		nb_place_libre--;
		lever_barriere(c);
	}
	void sortie() {
		unique_lock<mutex> l(m);
		nb_place_libre++;
		conditions[attentes_cotes.front()].notify_one();
	}
	void lever_barriere(cote c) {
		cout << "La barrière se lève pour laisser passer une voiture côté " << c << "\n";
	}
private:
	int nb_place_libre = NB_PLACES;
	mutex m;
	condition_variable conditions[NB_COTES];
	queue<cote> attentes_cotes;
};

Parking s;

void voiture(int n) {
	for (int i = 0; i < NB_ESSAIS; i++) {
		if (rand() % 2 == 0) usleep(100);
		if (rand() % 2 == 0) usleep(100);
		Parking::cote c = rand() % 2
			? Parking::cote::EST
			: Parking::cote::OUEST;
		printf("%d: entrer %d\n", n, c);
		s.demande_entree(c);
		printf("%d:       entré %d\n", n, c);
		usleep(10000);
		s.sortie();
		printf("%d:            sortie\n", n);
		if (rand() % 2 == 0) usleep(10000);
	}
}

int main() {
	srand(time(NULL));
	vector<thread> tab;
	for (int i = 0; i < NB_VOITURES; i++) {
		tab.push_back(thread(voiture, i));
	}
	for (thread &t: tab) {
		t.join();
	}
}
