#!/bin/sh

# Entree : un fichier tab-separated values exporté de TOMUSS avec des lignes :
# no-etudiant TAB ip_vm

if ! [ -f "./password.sh" ]; then
    cat >./password.sh <<EOF
SSHPASS=
SSHUSER=
EOF
fi

. ./password.sh

if [ -z "$SSHPASS" ]; then
    echo "Please set SSHPASS= in ./password.sh"
    exit 1
fi

if [ -z "$SSHUSER" ]; then
    echo "Please set SSHUSER= in ./password.sh"
    exit 1
fi
    
export SSHPASS

FILE=$(cat IP_VMs.txt | sed 's/\t/,/')

mv -f stats.tsv stats.tsv.bak
rm -f stats.log
cat > stats.tsv <<EOF
ID	Nom	Prénom	1	Connexion à la VM	1	Création de l'utilisateur lurong	1	Connexion en tant que lurong	1	Installation du paquet nginx	1	Installation du paquet links	1	Désinstallation du paquet apache2	1	Installation du paquet apache2	1	Début du jeu de piste	1	Fin du jeu de piste
			Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire	Note	Commentaire
EOF

pass() {
    msg=$1
    shift
    printf '1\t'"$msg"'\t' "$@"
}
fail() {
    msg=$1
    shift
    printf '0\t'"$msg"'\t' "$@"
}

password_changed() {
    if [ "$1" -gt 0 ]; then
	pass 'Le mot de passe a été changé sur la VM'
    else
	fail 'Le mot de passe n'\''a pas été changé sur la VM'
    fi
}

at_least_one_ctx() {
    if [ "$1" -gt 0 ]; then
	pass 'Il y a eu %s connexions de l'\''utilisateur %s sur la VM' "$1" "$2"
    else
	fail 'Pas de connexion de l'\''utilisateur %s sur la VM' "$2"
    fi
}

installed() {
    if [ "$1" -gt 0 ]; then
	pass 'Package %s installé sur la VM' "$2"
    else
	fail 'Package %s pas encore installé sur la VM' "$2"
    fi
}

not_installed() {
    if [ "$1" -gt 0 ]; then
	fail 'Package %s encore installé sur la VM' "$2"
    else
	pass 'Package %s jamais installé ou désinstallé sur la VM' "$2"
    fi
}

was_installed() {
    if [ "$1" -gt 0 ]; then
	pass 'Package %s a été installé sur la VM' "$2"
    else
	fail 'Package %s jamais installé sur la VM' "$2"
    fi
}

was_uninstalled() {
    if [ "$1" -gt 0 ]; then
	pass 'Package %s a été désinstallé sur la VM' "$2"
    else
	fail 'Package %s n'\''a pas été désinstallé sur la VM' "$2"
    fi
}

#FILE=11809142,192.168.75.254
for line in $FILE
do
    id=$(echo "$line" | cut -d , -f 1)
    vm=$(echo "$line" | cut -d , -f 2)
    echo $id '->' $vm
    if [ -z "$vm" ]; then
	continue
    fi

#    if SSHPASS=etudiant sshpass -e ssh -F/dev/null -oStrictHostKeyChecking=accept-new etudiant@$vm true
#    then
#	passwordchanged=0
#    else
#	passwordchanged=1
#    fi
    
    stats=$(
	sshpass -e ssh -F/dev/null -oStrictHostKeyChecking=accept-new $SSHUSER@$vm sh -c \
   'true; # Not sure why, in this context sh -c doesnt want an empty first line :-\
    # Feed password into sudo. Must be a binary, not a script.
    echo '\''int main() {puts("'"$SSHPASS"'");}'\''>/tmp/p.c;
    gcc /tmp/p.c -o /tmp/p -w;
    SUDO_ASKPASS=/tmp/p; export SUDO_ASKPASS;

    # logrotate empties wtmp and creates wtmp.* periodically ... search all wtmp* files
    wtmp=$(echo /var/log/wtmp* | sed "s@/var@-f /var@g")

    # Be future-proof, logrotate isn t helping us anyway
    sudo -A apt-get remove -q=2 logrotate >/dev/null
    # Just to avoid an error during next run
    sudo -A mkdir -p /home/autosave/

    check_install () {
    	printf ${2:-${1}}=;
        if dpkg --list | grep -q "^ii  $1 "; then echo 1; else echo 0; fi;
    }

    was_uninstalled () {
    	printf ${2:-${1}}=;
        if sudo -A grep -q "^Removing $1 " /var/log/apt/term.log || \
	   sudo -A zgrep -q "^Removing $1 " /var/log/apt/term.log*.gz; then echo 1; else echo 0; fi;
    }

    was_installed () {
    	printf ${2:-${1}}=;
        if sudo -A grep -q "^Setting up $1 " /var/log/apt/term.log || \
	   sudo -A zgrep -q "^Setting up $1 " /var/log/apt/term.log*.gz; then echo 1; else echo 0; fi;
    }

    # fetch data about progress

    printf etudiant=;
    last etudiant $wtmp | grep -c ^etudiant;

    printf nbluronglast=;
    last lurong $wtmp | grep -c ^lurong;

    printf nblurongsu=;
    sudo -A grep -c "\[[0-9]*\]: Successful su for lurong" /var/log/auth.log;

    printf lurong=;
    grep -c ^lurong /etc/passwd;

    was_installed nginx

    was_uninstalled apache2 apache2_wasuninstalled

    was_installed apache2 apache2_wasinstalled

    check_install links

    printf apachelogs=
    if test -d /var/log/apache2/; then echo 1; else echo 0; fi

    was_installed asr7

    was_installed asr7-web asr7web

'
	 )
    if printf '%s\n' "$stats" | grep -v -E '^[a-z0-9_]+=[0-9]+$'
    then
	# Potential security issue: we're going to feed this to
	# "eval", accept only basic assignments.
	echo "Internal error, stats not properly retrieved:"
	printf '%s\n' "$stats"
	exit 1
    fi

    printf '%s\n%s\n\n' "$line" "$stats" >> stats.log

    # Read values collected into shell variables
    eval "$stats"

    nblurong=$((nbluronglast + nblurongsu))
    (
	printf '%s\tnom-bidon\tprenom-bidon\t' "$id"
	at_least_one_ctx "$etudiant" etudiant
	#password_changed "$passwordchanged"
	if [ "$lurong" -gt 0 ]; then
	    pass 'Utilisateur lurong créé'
	else
	    fail 'Utilisateur lurong pas encore créé'
	fi
	at_least_one_ctx "$nblurong" lurong
	installed "$nginx" nginx
	installed "$links" links
	was_uninstalled "$apache2_wasuninstalled" apache2
	was_installed "$((apachelogs + apache2_wasinstalled))" apache2 
	if [ "$asr7" = 1 ]; then
	    pass "Jeu de piste démarré"
	else
	    fail "Première étape du jeu de piste non réalisée."
	fi
	if [ "$asr7web" = 1 ]; then
	    pass "Jeu de piste terminé (ou presque)"
	else
	    fail "Jeu de piste non-terminé."
	fi
	printf '\n'
    ) | tee -a stats.tsv
done
